@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2021 Derick Rethans                               |
+   | Copyright (c) 2002-2022 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -14,11 +14,15 @@
    +----------------------------------------------------------------------+
  */
 
-#include "php.h"
+#include "lib/php-header.h"
 #include "TSRM.h"
 #include "php_globals.h"
 #include "zend_closures.h"
 #include "zend_exceptions.h"
+#if PHP_VERSION_ID >= 80200
+# include "zend_attributes.h"
+#endif
+
 #include "zend_interfaces.h"
 
 #if PHP_VERSION_ID >= 80100
@@ -47,14 +51,10 @@ ZEND_EXTERN_MODULE_GLOBALS(xdebug)
 void (*xdebug_old_error_cb)(int type, zend_string *error_filename, const uint32_t error_lineno, zend_string *message);
 void (*xdebug_new_error_cb)(int type, zend_string *error_filename, const uint32_t error_lineno, zend_string *message);
 static void xdebug_error_cb(int orig_type, zend_string *error_filename, const uint32_t error_lineno, zend_string *message);
-#elif PHP_VERSION_ID >= 80000
+#else
 void (*xdebug_old_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, zend_string *message);
 void (*xdebug_new_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, zend_string *message);
 static void xdebug_error_cb(int orig_type, const char *error_filename, const uint32_t error_lineno, zend_string *message);
-#else
-void (*xdebug_old_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, const char *format, va_list args) ZEND_ATTRIBUTE_PTR_FORMAT(printf, 4, 0);
-void (*xdebug_new_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, const char *format, va_list args) ZEND_ATTRIBUTE_PTR_FORMAT(printf, 4, 0);
-static void xdebug_error_cb(int orig_type, const char *error_filename, const uint32_t error_lineno, const char *format, va_list args) ZEND_ATTRIBUTE_PTR_FORMAT(printf, 4, 0);
 #endif
 
 /* execution redirection functions */
@@ -334,12 +334,13 @@ normal_after_all:
 static void collect_params_internal(function_stack_entry *fse, zend_execute_data *zdata, zend_op_array *op_array)
 {
 	int i;
-	int is_variadic       = !!(zdata->func->common.fn_flags & ZEND_ACC_VARIADIC);
-	int is_trampoline     = !!(zdata->func->common.fn_flags & ZEND_ACC_CALL_VIA_TRAMPOLINE);
-	int variadic_at_pos   = NO_VARIADIC;
-	int names_expected    = 0;
-	int arguments_sent    = 0;
-	int arguments_storage = 0;
+	int is_variadic        = !!(zdata->func->common.fn_flags & ZEND_ACC_VARIADIC);
+	int is_trampoline      = !!(zdata->func->common.fn_flags & ZEND_ACC_CALL_VIA_TRAMPOLINE);
+	int variadic_at_pos    = NO_VARIADIC;
+	int variadic_sensitive = 0;
+	int names_expected     = 0;
+	int arguments_sent     = 0;
+	int arguments_storage  = 0;
 
 	arguments_sent = ZEND_CALL_NUM_ARGS(zdata);
 	if (arguments_sent > USHRT_MAX) {
@@ -402,18 +403,41 @@ static void collect_params_internal(function_stack_entry *fse, zend_execute_data
 
 	/* Collect Arguments */
 	for (i = 0; i < arguments_sent; i++) {
+#if PHP_VERSION_ID >= 80200
+		zend_attribute *attribute;
+#else
+		void *attribute = NULL;
+#endif
+
 		/* The index in ZEND_CALL_ARG is 1-based */
 #if DEBUG
-		fprintf(stderr, "Copying argument %d: ", i);
-		fprintf(stderr, "ARG ");
+		fprintf(stderr, "Copying argument %d\n", i);
 #endif
-		ZVAL_COPY(&(fse->var[i].data), ZEND_CALL_ARG(zdata, i + 1));
+#if PHP_VERSION_ID >= 80200
+		attribute = zend_get_parameter_attribute_str(
+			zdata->func->common.attributes,
+			"sensitiveparameter",
+			sizeof("sensitiveparameter") - 1,
+			i
+		);
+#endif
+		if (attribute && fse->var[i].is_variadic) {
+			variadic_sensitive = 1;
+		}
+# if DEBUG
+		fprintf(stderr, "SENSTIVIVE %d ", attribute != NULL);
+# endif
+
+		if ((variadic_sensitive || attribute != NULL) && !fse->var[i].is_variadic) {
+			ZVAL_STRING(&(fse->var[i].data), "[Sensitive Parameter]");
+		} else {
+			ZVAL_COPY(&(fse->var[i].data), ZEND_CALL_ARG(zdata, i + 1));
+		}
 #if DEBUG
 		fprintf(stderr, "OK\n");
 #endif
 	}
 
-#if PHP_VERSION_ID >= 80000
 	if (ZEND_CALL_INFO(zdata) & ZEND_CALL_HAS_EXTRA_NAMED_PARAMS) {
 		zend_string *name;
 		zval        *param;
@@ -429,7 +453,6 @@ static void collect_params_internal(function_stack_entry *fse, zend_execute_data
 			i++;
 		} ZEND_HASH_FOREACH_END();
 	}
-#endif
 
 #if DEBUG
 	for (i = 0; i < fse->varc; i++) {
@@ -441,12 +464,13 @@ static void collect_params_internal(function_stack_entry *fse, zend_execute_data
 static void collect_params(function_stack_entry *fse, zend_execute_data *zdata, zend_op_array *op_array)
 {
 	int i;
-	int is_variadic       = !!(zdata->func->common.fn_flags & ZEND_ACC_VARIADIC);
-	int is_trampoline     = !!(zdata->func->common.fn_flags & ZEND_ACC_CALL_VIA_TRAMPOLINE);
-	int variadic_at_pos   = NO_VARIADIC;
-	int names_expected    = 0;
-	int arguments_sent    = 0;
-	int arguments_storage = 0;
+	int is_variadic        = !!(zdata->func->common.fn_flags & ZEND_ACC_VARIADIC);
+	int is_trampoline      = !!(zdata->func->common.fn_flags & ZEND_ACC_CALL_VIA_TRAMPOLINE);
+	int variadic_at_pos    = NO_VARIADIC;
+	int variadic_sensitive = 0;
+	int names_expected     = 0;
+	int arguments_sent     = 0;
+	int arguments_storage  = 0;
 
 	arguments_sent = ZEND_CALL_NUM_ARGS(zdata);
 
@@ -502,6 +526,12 @@ static void collect_params(function_stack_entry *fse, zend_execute_data *zdata,
 
 	/* Collect Arguments */
 	for (i = 0; i < fse->varc; i++) {
+#if PHP_VERSION_ID >= 80200
+		zend_attribute *attribute;
+#else
+		void *attribute = NULL;
+#endif
+
 		/* The index in ZEND_CALL_ARG is 1-based */
 #if DEBUG
 		fprintf(stderr, "Copying argument %d: ", i);
@@ -510,19 +540,41 @@ static void collect_params(function_stack_entry *fse, zend_execute_data *zdata,
 #if DEBUG
 			fprintf(stderr, "ARG ");
 #endif
-			ZVAL_COPY(&(fse->var[i].data), ZEND_CALL_ARG(zdata, i + 1));
+#if PHP_VERSION_ID >= 80200
+			attribute = zend_get_parameter_attribute_str(
+				zdata->func->common.attributes,
+				"sensitiveparameter",
+				sizeof("sensitiveparameter") - 1,
+				i
+			);
+#endif
+			if (attribute && fse->var[i].is_variadic) {
+				variadic_sensitive = 1;
+			}
+# if DEBUG
+			fprintf(stderr, "SENSITIVE %d ", attribute != NULL);
+# endif
+
+			if ((variadic_sensitive || attribute != NULL) && !fse->var[i].is_variadic) {
+				ZVAL_STRING(&(fse->var[i].data), "[Sensitive Parameter]");
+			} else {
+				ZVAL_COPY(&(fse->var[i].data), ZEND_CALL_ARG(zdata, i + 1));
+			}
 		} else {
 #if DEBUG
 			fprintf(stderr, "VAR_NUM ");
 #endif
-			ZVAL_COPY(&(fse->var[i].data), ZEND_CALL_VAR_NUM(zdata, zdata->func->op_array.last_var + zdata->func->op_array.T + i - names_expected));
+			if (variadic_sensitive) {
+				ZVAL_STRING(&(fse->var[i].data), "[Sensitive Parameter]");
+			} else {
+				ZVAL_COPY(&(fse->var[i].data), ZEND_CALL_VAR_NUM(zdata, zdata->func->op_array.last_var + zdata->func->op_array.T + i - names_expected));
+			}
 		}
 #if DEBUG
 		fprintf(stderr, "OK\n");
 #endif
 	}
 
-#if PHP_VERSION_ID >= 80000
 	if (ZEND_CALL_INFO(zdata) & ZEND_CALL_HAS_EXTRA_NAMED_PARAMS) {
 		zend_string *name;
 		zval        *param;
@@ -538,7 +590,6 @@ static void collect_params(function_stack_entry *fse, zend_execute_data *zdata,
 			i++;
 		} ZEND_HASH_FOREACH_END();
 	}
-#endif
 
 #if DEBUG
 	for (i = 0; i < fse->varc; i++) {
@@ -636,6 +687,7 @@ function_stack_entry *xdebug_add_stack_frame(zend_execute_data *zdata, zend_op_a
 	} else {
 		tmp->lineno = find_line_number_for_current_execute_point(edata);
 		tmp->is_variadic = !!(zdata->func->common.fn_flags & ZEND_ACC_VARIADIC);
+		tmp->is_trampoline = !!(zdata->func->common.fn_flags & ZEND_ACC_CALL_VIA_TRAMPOLINE);
 
 		if (XDEBUG_MODE_IS(XDEBUG_MODE_TRACING) || XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
 			if (ZEND_USER_CODE(zdata->func->type)) {
@@ -769,7 +821,7 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 		}
 
 		/* Check for entry breakpoints */
-		xdebug_debugger_handle_breakpoints(fse, XDEBUG_BREAKPOINT_TYPE_CALL);
+		xdebug_debugger_handle_breakpoints(fse, XDEBUG_BREAKPOINT_TYPE_CALL|XDEBUG_BREAKPOINT_TYPE_EXTERNAL, NULL);
 	}
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_PROFILING)) {
@@ -795,8 +847,14 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 	}
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
+		zval *return_value = NULL;
+
+		if (!fse->is_trampoline && execute_data->return_value && !(op_array->fn_flags & ZEND_ACC_GENERATOR)) {
+			return_value = execute_data->return_value;
+		}
+
 		/* Check for return breakpoints */
-		xdebug_debugger_handle_breakpoints(fse, XDEBUG_BREAKPOINT_TYPE_RETURN);
+		xdebug_debugger_handle_breakpoints(fse, XDEBUG_BREAKPOINT_TYPE_RETURN|XDEBUG_BREAKPOINT_TYPE_EXTERNAL, return_value);
 	}
 
 	fse->symbol_table = NULL;
@@ -843,10 +901,8 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 	int                   restore_error_handler_situation = 0;
 #if PHP_VERSION_ID >= 80100
 	void                (*tmp_error_cb)(int type, zend_string *error_filename, const uint32_t error_lineno, zend_string *message) = NULL;
-#elif PHP_VERSION_ID >= 80000
-	void                (*tmp_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, zend_string *message) = NULL;
 #else
-	void                (*tmp_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, const char *format, va_list args) ZEND_ATTRIBUTE_PTR_FORMAT(printf, 4, 0) = NULL;
+	void                (*tmp_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, zend_string *message) = NULL;
 #endif
 
 	/* If the stack vector hasn't been initialised yet, we should abort immediately */
@@ -877,7 +933,7 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
 		/* Check for entry breakpoints */
-		xdebug_debugger_handle_breakpoints(fse, XDEBUG_BREAKPOINT_TYPE_CALL);
+		xdebug_debugger_handle_breakpoints(fse, XDEBUG_BREAKPOINT_TYPE_CALL, NULL);
 	}
 
 	/* Check for SOAP */
@@ -919,7 +975,7 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
 		/* Check for return breakpoints */
-		xdebug_debugger_handle_breakpoints(fse, XDEBUG_BREAKPOINT_TYPE_RETURN);
+		xdebug_debugger_handle_breakpoints(fse, XDEBUG_BREAKPOINT_TYPE_RETURN, return_value);
 	}
 
 	if (XG_BASE(stack)) {
@@ -1043,12 +1099,22 @@ static void add_fiber_main(zend_fiber_context *fiber)
 	xdebug_str           *key = create_key_for_fiber(fiber);
 	function_stack_entry *tmp = (function_stack_entry*) xdebug_vector_push(XG_BASE(stack));
 
-	tmp->level = 0;
+	tmp->level        = XDEBUG_VECTOR_COUNT(XG_BASE(stack));
 	tmp->user_defined = XDEBUG_BUILT_IN;
 	tmp->function.type = XFUNC_FIBER;
 	tmp->function.object_class = NULL;
 	tmp->function.scope_class = NULL;
 	tmp->function.function = xdstrdup(key->d);
+	tmp->filename = zend_string_copy(zend_get_executed_filename_ex());
+	tmp->lineno = zend_get_executed_lineno();
+
+	tmp->prev_memory = XG_BASE(prev_memory);
+	tmp->memory = zend_memory_usage(0);
+	XG_BASE(prev_memory) = tmp->memory;
+
+	tmp->nanotime = xdebug_get_nanotime();
+
+	xdebug_str_free(key);
 }
 
 static xdebug_vector* create_stack_for_fiber(zend_fiber_context *fiber)
@@ -1105,6 +1171,69 @@ static void xdebug_fiber_switch_observer(zend_fiber_context *from, zend_fiber_co
 }
 #endif
 
+#ifdef __linux__
+int read_systemd_private_tmp_directory(char **private_tmp)
+{
+	pid_t       current_pid;
+	char       *mountinfo_fn;
+	FILE       *mountinfo_fd;
+	size_t      bytes_read;
+	char        buffer[8192] = { 0 };
+	xdebug_arg *lines;
+	int         i;
+	int         retval = 0;
+
+	/* Open right file in /proc */
+	current_pid = getpid();
+	mountinfo_fn = xdebug_sprintf("/proc/%ld/mountinfo", current_pid);
+	mountinfo_fd = fopen(mountinfo_fn, "r");
+	xdfree(mountinfo_fn);
+	if (!mountinfo_fd) {
+		return retval;
+	}
+
+	/* Read contents and split in lines */
+	bytes_read = fread(buffer, 1, sizeof(buffer), mountinfo_fd);
+	lines = xdebug_arg_ctor();
+	xdebug_explode("\n", buffer, lines, -1);
+
+	/* Check whether each line has /tmp/systemd-private, and parse accordingly.
+	 * There is a " " in front as there is often also a /var/tmp/systemd-private
+	 * entry that we need to ignore. */
+	for (i = 0; i < lines->c; i++) {
+		const char *mountpoint;
+		const char *slash;
+
+		mountpoint = strstr(lines->args[i], " /tmp/systemd-private");
+		if (mountpoint == NULL) {
+			continue;
+		}
+
+		mountpoint++;
+
+		slash = strchr(mountpoint + 1, '/');
+		if (!slash) {
+			continue;
+		}
+
+		slash = strchr(slash + 1, '/');
+		if (!slash) {
+			continue;
+		}
+
+		*private_tmp = xdstrndup(mountpoint, slash - mountpoint);
+
+		retval = 1;
+		break;
+	}
+
+	/* Clean up and return */
+	xdebug_arg_dtor(lines);
+	fclose(mountinfo_fd);
+	return retval;
+}
+#endif
+
 void xdebug_base_minit(INIT_FUNC_ARGS)
 {
 	/* Record Zend and Xdebug error callbacks, the actual setting is done in
@@ -1112,13 +1241,6 @@ void xdebug_base_minit(INIT_FUNC_ARGS)
 	xdebug_old_error_cb = zend_error_cb;
 	xdebug_new_error_cb = xdebug_error_cb;
 
-	/* Redirect compile and execute functions to our own. For PHP 7.3 and
-	 * later, we hook these in xdebug_post_startup instead */
-#if PHP_VERSION_ID < 70300
-	old_compile_file = zend_compile_file;
-	zend_compile_file = xdebug_compile_file;
-#endif
-
 	xdebug_old_execute_ex = zend_execute_ex;
 	zend_execute_ex = xdebug_execute_ex;
 
@@ -1132,6 +1254,11 @@ void xdebug_base_minit(INIT_FUNC_ARGS)
 #if PHP_VERSION_ID >= 80100
 	zend_observer_fiber_switch_register(xdebug_fiber_switch_observer);
 #endif
+
+	XG_BASE(private_tmp) = NULL;
+#ifdef __linux__
+	read_systemd_private_tmp_directory(&XG_BASE(private_tmp));
+#endif
 }
 
 void xdebug_base_mshutdown()
@@ -1141,6 +1268,12 @@ void xdebug_base_mshutdown()
 	zend_execute_ex = xdebug_old_execute_ex;
 	zend_execute_internal = xdebug_old_execute_internal;
 	zend_error_cb = xdebug_old_error_cb;
+
+#ifdef __linux__
+	if (XG_BASE(private_tmp)) {
+		xdfree(XG_BASE(private_tmp));
+	}
+#endif
 }
 
 void xdebug_base_post_startup()
@@ -1196,6 +1329,10 @@ void xdebug_base_rinit()
 	XG_BASE(filters_tracing)           = xdebug_llist_alloc(xdebug_llist_string_dtor);
 
 	xdebug_base_overloaded_functions_setup();
+
+	if (XG_BASE(private_tmp)) {
+		xdebug_log_ex(XLOG_CHAN_CONFIG, XLOG_INFO, "PRIVTMP", "Systemd Private Temp Directory is enabled (%s)", XG_BASE(private_tmp));
+	}
 }
 
 void xdebug_base_post_deactivate()
@@ -1253,7 +1390,7 @@ static void xdebug_error_cb(int orig_type, zend_string *error_filename, const un
 		xdebug_old_error_cb(orig_type, error_filename, error_lineno, message);
 	}
 }
-#elif PHP_VERSION_ID >= 80000
+#else
 static void xdebug_error_cb(int orig_type, const char *error_filename, const unsigned int error_lineno, zend_string *message)
 {
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
@@ -1272,32 +1409,6 @@ static void xdebug_error_cb(int orig_type, const char *error_filename, const uns
 		xdebug_old_error_cb(orig_type, error_filename, error_lineno, message);
 	}
 }
-#else
-static void xdebug_error_cb(int orig_type, const char *error_filename, const unsigned int error_lineno, const char *format, va_list args)
-{
-	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
-		int          type               = orig_type & E_ALL;
-		char        *error_type_str     = xdebug_error_type(type);
-		zend_string *tmp_error_filename = zend_string_init(error_filename, strlen(error_filename), 0);
-		char        *buffer;
-		va_list      new_args;
-
-		va_copy(new_args, args);
-		vspprintf(&buffer, PG(log_errors_max_len), format, new_args);
-		va_end(new_args);
-
-		xdebug_debugger_error_cb(tmp_error_filename, error_lineno, type, error_type_str, buffer);
-
-		efree(buffer);
-		zend_string_release(tmp_error_filename);
-		xdfree(error_type_str);
-	}
-	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
-		xdebug_develop_error_cb(orig_type, error_filename, error_lineno, format, args);
-	} else {
-		xdebug_old_error_cb(orig_type, error_filename, error_lineno, format, args);
-	}
-}
 #endif
 
 void xdebug_base_use_original_error_cb(void)
@@ -1310,13 +1421,8 @@ void xdebug_base_use_xdebug_error_cb(void)
 	zend_error_cb = xdebug_new_error_cb;
 }
 
-#if PHP_VERSION_ID >= 80000
 static void xdebug_throw_exception_hook(zend_object *exception)
 {
-#else
-static void xdebug_throw_exception_hook(zval *exception)
-{
-#endif
 	zval *code, *message, *file, *line;
 	zend_class_entry *exception_ce;
 	char *code_str = NULL;
@@ -1330,12 +1436,18 @@ static void xdebug_throw_exception_hook(zval *exception)
 		return;
 	}
 
-#if PHP_VERSION_ID >= 80000
-	exception_ce = exception->ce;
-#else
-	exception_ce = Z_OBJCE_P(exception);
+	if (zend_is_unwind_exit(exception)) {
+		return;
+	}
+
+#if PHP_VERSION_ID >= 80100
+	if (zend_is_graceful_exit(exception)) {
+		return;
+	}
 #endif
 
+	exception_ce = exception->ce;
+
 	code =    zend_read_property(exception_ce, exception, "code",    sizeof("code")-1,    0, &dummy);
 	message = zend_read_property(exception_ce, exception, "message", sizeof("message")-1, 0, &dummy);
 	file =    zend_read_property(exception_ce, exception, "file",    sizeof("file")-1,    0, &dummy);
