@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2021 Derick Rethans                               |
+   | Copyright (c) 2002-2023 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -183,15 +183,7 @@ static int object_item_add_zend_prop_to_merged_hash(zend_property_info *zpp, Has
 
 	item = xdmalloc(sizeof(xdebug_object_item));
 	item->type = object_type;
-#if ZTS && PHP_VERSION_ID < 70400
-	if (ce->type == 1) {
-		item->zv   = &CG(static_members_table)[(zend_intptr_t) CE_STATIC_MEMBERS(ce)][zpp->offset];
-	} else {
-		item->zv   = &CE_STATIC_MEMBERS(ce)[zpp->offset];
-	}
-#else
 	item->zv   = &CE_STATIC_MEMBERS(ce)[zpp->offset];
-#endif
 	item->name = (char*) STR_NAME_VAL(zpp->name);
 	item->name_len = STR_NAME_LEN(zpp->name);
 
@@ -405,15 +397,14 @@ static int xdebug_object_element_export_xml_node(xdebug_object_item *item_nptr,
 		}
 
 
-		xdebug_xml_add_attribute_ex(
-			node, "facet",
-			xdebug_sprintf("%s%s%s",
-				(*item)->type & XDEBUG_OBJECT_ITEM_TYPE_STATIC_PROPERTY ? "static " : "",
-				modifier,
-				(*item)->type & XDEBUG_OBJECT_ITEM_TYPE_READONLY ? " readonly" : ""
-			),
-			0, 1
-		);
+		if ((*item)->type & XDEBUG_OBJECT_ITEM_TYPE_STATIC_PROPERTY) {
+			xdebug_xml_expand_attribute_value(node, "facet", "static");
+		}
+		xdebug_xml_expand_attribute_value(node, "facet", modifier);
+		if ((*item)->type & XDEBUG_OBJECT_ITEM_TYPE_READONLY) {
+			xdebug_xml_expand_attribute_value(node, "facet", "readonly");
+		}
+
 		xdebug_xml_add_child(parent, node);
 		xdebug_var_export_xml_node(&((*item)->zv), tmp_fullname ? tmp_fullname : NULL, node, options, level + 1);
 
@@ -461,7 +452,8 @@ static void xdebug_var_xml_attach_property_with_contents(zend_property_info *pro
 	xdfree(prop_class_name);
 
 	if (contents) {
-		xdebug_xml_add_attribute_ex(contents, "facet", xdebug_sprintf("static %s", modifier), 0, 1);
+		xdebug_xml_expand_attribute_value(contents, "facet", "static");
+		xdebug_xml_expand_attribute_value(contents, "facet", modifier);
 		xdebug_xml_add_child(node, contents);
 	} else {
 		xdebug_var_xml_attach_uninitialized_var(options, node, xdebug_str_create(ZSTR_VAL(prop_info->name), ZSTR_LEN(prop_info->name)));
@@ -544,7 +536,7 @@ void xdebug_var_export_xml_node(zval **struc, xdebug_str *name, xdebug_xml_node
 
 		case IS_DOUBLE:
 			xdebug_xml_add_attribute(node, "type", "float");
-			add_unencoded_text_value_attribute_or_element(options, node, xdebug_sprintf("%.*G", (int) EG(precision), Z_DVAL_P(*struc)));
+			add_unencoded_text_value_attribute_or_element(options, node, xdebug_sprintf("%.*H", (int) EG(precision), Z_DVAL_P(*struc)));
 			break;
 
 		case IS_STRING:
@@ -593,9 +585,6 @@ void xdebug_var_export_xml_node(zval **struc, xdebug_str *name, xdebug_xml_node
 			HashTable          *merged_hash;
 			zend_string        *class_name;
 			zend_class_entry   *ce;
-#if PHP_VERSION_ID < 70400
-			int                 is_temp;
-#endif
 			int                 extra_children = 0;
 			zend_property_info *zpi_val;
 
@@ -610,7 +599,7 @@ void xdebug_var_export_xml_node(zval **struc, xdebug_str *name, xdebug_xml_node
 
 #if PHP_VERSION_ID >= 80100
 			zend_class_init_statics(ce);
-#elif PHP_VERSION_ID >= 70400
+#else
 			if (ce->type == ZEND_INTERNAL_CLASS || (ce->ce_flags & ZEND_ACC_IMMUTABLE)) {
 				zend_class_init_statics(ce);
 			}
@@ -623,11 +612,8 @@ void xdebug_var_export_xml_node(zval **struc, xdebug_str *name, xdebug_xml_node
 			xdebug_zend_hash_apply_protection_end(&ce->properties_info);
 
 			/* Adding normal properties */
-#if PHP_VERSION_ID >= 70400
-			myht = xdebug_objdebug_pp(struc);
-#else
-			myht = xdebug_objdebug_pp(struc, &is_temp);
-#endif
+			myht = xdebug_objdebug_pp(struc, XDEBUG_VAR_OBJDEBUG_DEFAULT);
+
 			if (myht) {
 				zval *tmp_val;
 
@@ -637,7 +623,7 @@ void xdebug_var_export_xml_node(zval **struc, xdebug_str *name, xdebug_xml_node
 					int flags = XDEBUG_OBJECT_ITEM_TYPE_PROPERTY;
 
 #if PHP_VERSION_ID >= 80100
-					if (ce->type != ZEND_INTERNAL_CLASS) {
+					if (ce->type != ZEND_INTERNAL_CLASS && !HASH_KEY_IS_NUMERIC(key)) {
 						const char         *cls_name, *tmp_prop_name;
 						size_t              tmp_prop_name_len;
 						zend_string        *unmangled;
@@ -673,14 +659,9 @@ void xdebug_var_export_xml_node(zval **struc, xdebug_str *name, xdebug_xml_node
 #endif
 
 			if (instanceof_function(Z_OBJCE_P(*struc), zend_ce_closure)) {
+#if PHP_VERSION_ID < 80200
 				xdebug_xml_node *closure_cont, *closure_func;
-#if PHP_VERSION_ID >= 80000
 				const zend_function *closure_function = zend_get_closure_method_def(Z_OBJ_P(*struc));
-#else
-				const zend_function *closure_function = zend_get_closure_method_def(*struc);
-#endif
-
-				xdebug_xml_expand_attribute_value(node, "facet", "closure");
 
 				closure_cont = xdebug_xml_node_init("property");
 				xdebug_xml_add_attribute(closure_cont, "facet", "virtual readonly");
@@ -727,6 +708,9 @@ void xdebug_var_export_xml_node(zval **struc, xdebug_str *name, xdebug_xml_node
 
 				xdebug_xml_add_child(node, closure_cont);
 				extra_children = 1;
+#endif
+
+				xdebug_xml_expand_attribute_value(node, "facet", "closure");
 			}
 
 			{
@@ -735,7 +719,7 @@ void xdebug_var_export_xml_node(zval **struc, xdebug_str *name, xdebug_xml_node
 				tmp_str.l = ZSTR_LEN(class_name);
 				add_xml_attribute_or_element(options, node, "classname", 9, &tmp_str);
 			}
-			xdebug_xml_add_attribute(node, "children", merged_hash->nNumOfElements ? "1" : "0");
+			xdebug_xml_add_attribute(node, "children", (merged_hash->nNumOfElements || extra_children) ? "1" : "0");
 
 
 			if (!myht || !xdebug_zend_hash_is_recursive(myht)) {
@@ -768,11 +752,8 @@ void xdebug_var_export_xml_node(zval **struc, xdebug_str *name, xdebug_xml_node
 
 			zend_hash_destroy(merged_hash);
 			FREE_HASHTABLE(merged_hash);
-#if PHP_VERSION_ID >= 70400
 			zend_release_properties(myht);
-#else
-			xdebug_var_maybe_destroy_ht(myht, is_temp);
-#endif
+
 			break;
 		}
 
